/* rmap - vector based global map generating program
 * Copyright (C) 2000 Reza Naima <reza@reza.net>
 *
 * http://www.reza.net/rmap/
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "rmap.h"
#include "cities.h"
#include "rgb.h"

//The following subroutine is the only one borrowed from 
//the original mapping code -- strait forward bit flipping stuff
//which was contributed by Joe Dellinger (no email for me to
//contact him) 
//
// If you have a better way of accomplishing this, please
// Let me know as I'm sure this can't be supper efficient

#ifdef WORDS_BIGENDIAN
bitflip (BIT32 *x) {
	BIT32 y=0;
	
	y |= 0xFF000000 & ((0x000000FF & *x) << 24);
	y |= 0x00FF0000 & ((0x0000FF00 & *x) << 8);
	y |= 0x0000FF00 & ((0x00FF0000 & *x) >> 8);
	y |= 0x000000FF & ((0xFF000000 & *x) >> 24);
	*x = y;
}
#endif /* WORDS_BIGENDIAN */


void lc(char *str) {
	// this assumes a null terminating string
	int i; 
	for (i=0; str[i]; i++) 
		str[i] = tolower(str[i]);
	
}

// return a pointer to the color struct that 
// contains the rgb values for a color
// or null if nothing is found that matches
struct color* get_color (char* c) {
	int i=0;
	lc(c);
	
	while (colors[i].name != NULL) {
		if (!strcmp(colors[i].name,c))
			return &colors[i];
		i++;
	}
	return NULL;
}


void parse_colorfile (struct imageLayout *s) {
	FILE *f;
	struct color *c;
	int i,j,pos=0;
	int ret, line_number=0;
	int cat, type;
	char string[256];
	char color_name[100];

	// initialize stuff	
	bzero(string,256);

	// read in the color config file and allocate colors
	f = fopen(s->o.colorfile, "r");
	if (f == NULL) {
		fprintf(stderr,"Couldn't open %s (colormap file)\n", s->o.colorfile);
		exit(-2);
	}

	//now parse the file
	while ((ret = fread(string, 1, 255, f)) != 0) {
		line_number++;
		pos = (int) ftell(f) - ret;
		for (i=0; i<255; i++) {
			if (string[i] == '\n') {
				string[i] = '\0';
				fseek(f,pos+i+1, SEEK_SET);
				break;
			}
		}
		for (j=0; j<i; j++) {
			if (string[j] == '#')  {
				string[j] = '\0';
				break;
			}
		} 
		while (j>0) {
			j--;
			if (string[j] == ' ' || string [j] == '\t') {
				string[j] = '\0';
			} else {
				break;
			}
		}
		if (strlen(string) == 0) 
			continue;

		ret = sscanf(string, "%d %d %s", &cat, &type, color_name);
		if (ret != 3) {
			fprintf(stderr,"Unable to parse the string [%s] on line %d in %s\n", 
				string,line_number,s->o.colorfile);	
			continue;
		}
		if (s->colormap[cat][type] != NOPRINT) {
			// if we allocated a color somewere else, then 
			// that should overrride the one in the colormapfile
			continue;
		}
		if (cat > 4 || type > 15 || cat < 0 || type < 0) {
			fprintf(stderr,"Declaration value out of range (%d,%d) on line %d in %s\n",
				cat,type, line_number, s->o.colorfile);
		} else {
			if (!strcmp(color_name, "none")) {
				s->colormap[cat][type] = NOPRINT;
			} else {	
				c = get_color(color_name);
				if (c == NULL) {
					fprintf(stderr,"The color %s was not found on line %d in %s\n",
						color_name,line_number,s->o.colorfile);
					continue;
				} else {
					i = s->colormap[cat][type] = gdImageColorAllocate(s->im, c->r, c->g, c->b);
					s->colorunmap[i] = c;
				}
			}
		}
	}
	fclose(f);
}

void image_init (struct imageLayout *s) {
	int i,j;

	s->im = gdImageCreate(s->width, s->height);

	// default is not to print anthing ;)
	for (i=0; i< 5; i++)
		for (j=0; j<16; j++)
			s->colormap[i][j] = NOPRINT; //no print is default
}


void drawLine (struct imageLayout *s, vector a, vector b, int color_index) {
  	int x1,x2,y1,y2;
		
        //lets try it witout perspective
        x1 = a.item[X] + s->width/2;
        y1 = a.item[Y] + s->height/2;
        x2 = b.item[X] + s->width/2;
        y2 = b.item[Y] + s->height/2;


        // draw the point
        gdImageLine(s->im, x1,y1, x2,y2, color_index);
}   

void drawPoint (struct imageLayout *s, vector a, int color_index) {
	int x,y;

	x = a.item[X] + s->width/2;
	y = a.item[Y] + s->height/2;

	gdImageSetPixel(s->im, x, y, color_index);
}


void makePicture (struct imageLayout *s) {
	FILE *f;

	/* Default is to make the picture interlaced */	
	gdImageInterlace(s->im, 1);	

	f = fopen(s->o.outfile, "wb");
	if (f  == NULL) {
		fprintf(stderr, "Unable to open output file %s\n", s->o.outfile);
		return;
	}

/* even though gd is cool, they're fucking' lame
** for removing gif support */
#ifdef GD_GIF
	gdImageGif(s->im, f);
#else
#ifdef GD_PNG
	gdImagePng(s->im, f);
#else 
#ifdef GD_JPEG
	gdImageJpeg(s->im, f, 95);
#else 
#error "Your GD lib doesn't support GIF, PNG, or JPEG.  Verify that libgd is installed properly."
#endif /* GD_JPEG */
#endif /* GD_PNG */
#endif /* GD_GIF */

	fclose(f);
}

static matrix rotateZ (double deg) {
	matrix m;
	double d;

	d = CONV * deg;

	m.item[0][0] =  cos(d);
	m.item[0][1] =  sin(d);
	m.item[0][2] =  0.;
	m.item[1][0] = -sin(d);
	m.item[1][1] =  cos(d);
	m.item[1][2] =  0.;
	m.item[2][0] =  0.;
	m.item[2][1] =  0.;
	m.item[2][2] =  1.;
	
	return m;
}
static matrix rotateX (double deg) {
	matrix m;
	double d;

	d = CONV * deg;

	m.item[0][0] =  1.;
	m.item[0][1] =  0.;
	m.item[0][2] =  0.;
	m.item[1][0] =  0.;
	m.item[1][1] =  cos(d);
	m.item[1][2] = -sin(d);
	m.item[2][0] =  0.;
	m.item[2][1] =  sin(d);
	m.item[2][2] =  cos(d);
	
	return m;
}
static matrix rotateY (double deg) {
	matrix m;
	double d;

	d = CONV * deg;

	m.item[0][0] =  cos(d);
	m.item[0][1] =  0.;
	m.item[0][2] =  sin(d);
	m.item[1][0] =  0.;
	m.item[1][1] =  1.;
	m.item[1][2] =  0.;
	m.item[2][0] = -sin(d);
	m.item[2][1] =  0.;
	m.item[2][2] =  cos(d);
	
	return m;
}

static vector multiply23 (vector a, matrix b) {
	vector v;
	
	v.item[0] = a.item[0]*b.item[0][0] + a.item[1]*b.item[1][0] + a.item[2]*b.item[2][0];
	v.item[1] = a.item[0]*b.item[0][1] + a.item[1]*b.item[1][1] + a.item[2]*b.item[2][1];
	v.item[2] = a.item[0]*b.item[0][2] + a.item[1]*b.item[1][2] + a.item[2]*b.item[2][2];

	return v;
}

static matrix multiply33 (matrix a, matrix b) {
	matrix m;

	m.item[0][0] = a.item[0][0]*b.item[0][0] + a.item[0][1]*b.item[1][0] + a.item[0][2]*b.item[2][0];
	m.item[0][1] = a.item[0][0]*b.item[0][1] + a.item[0][1]*b.item[1][1] + a.item[0][2]*b.item[2][1];
	m.item[0][2] = a.item[0][0]*b.item[0][2] + a.item[0][1]*b.item[1][2] + a.item[0][2]*b.item[2][2];
	m.item[1][0] = a.item[1][0]*b.item[0][0] + a.item[1][1]*b.item[1][0] + a.item[1][2]*b.item[2][0];
	m.item[1][1] = a.item[1][0]*b.item[0][1] + a.item[1][1]*b.item[1][1] + a.item[1][2]*b.item[2][1];
	m.item[1][2] = a.item[1][0]*b.item[0][2] + a.item[1][1]*b.item[1][2] + a.item[1][2]*b.item[2][2];
	m.item[2][0] = a.item[2][0]*b.item[0][0] + a.item[2][1]*b.item[1][0] + a.item[2][2]*b.item[2][0];
	m.item[2][1] = a.item[2][0]*b.item[0][1] + a.item[2][1]*b.item[1][1] + a.item[2][2]*b.item[2][1];
	m.item[2][2] = a.item[2][0]*b.item[0][2] + a.item[2][1]*b.item[1][2] + a.item[2][2]*b.item[2][2];
	
	return m;
}


static vector sphere2cart (spherePt s) {
        vector v;

        v.item[X] = s.item[RADIUS] * cos(s.item[LATITUDE] * CONV) * sin(-1*s.item[LONGITUDE] * CONV);
        v.item[Y] = s.item[RADIUS] * sin(s.item[LATITUDE] * CONV);
        v.item[Z] = s.item[RADIUS] * cos(s.item[LATITUDE] * CONV) * cos(-1*s.item[LONGITUDE] * CONV);
        return v;
}

static matrix scale (double x, double y, double z) {
	matrix m;
	bzero(&m, sizeof(m));

	m.item[X][X] = x;
	m.item[Y][Y] = y;
	m.item[Z][Z] = z;

	return m;
}

	

static int segment_in_scene (struct imageLayout *s, struct segment_index *i) {
	spherePt max, min;
	vector pt1, pt2;
	int bit_number[9] = {0,1,2,0,3,0,0,0,4};

	
	// first verify if we want this printed
	if (s->colormap[bit_number[i->category]][i->type] == NOPRINT)
		return 0;

	max.item[LONGITUDE] = i->maxlong/3600.;
	max.item[LATITUDE] = i->maxlat/3600.;
	max.item[RADIUS] = EARTH_RADIUS;

	min.item[LONGITUDE] = i->minlong/3600.;
	min.item[LATITUDE] = i->minlat/3600.;
	min.item[RADIUS] = EARTH_RADIUS;

	pt1 = multiply23(sphere2cart(max), s->rotation);
	pt2 = multiply23(sphere2cart(min), s->rotation);

	// check to see if it is z-clipped
	if (s->transparent == 0 && (pt1.item[Z] < 0 || pt2.item[Z] < 0)) {
		return 0; 
	}
		
	/* Yeah, this is ugly, but my head hurts right now. I've got the flu
	   I'll try to figure it out later */

	if (    (pt1.item[X] > s->width/2       && pt2.item[X] > s->width/2)          ||
		(pt1.item[Y] > s->width/2       && pt2.item[Y] > s->width/2)          || 
		(pt1.item[X] < (-1.*s->width/2) && pt2.item[X] < (-1.*s->width/2) )   ||
		(pt1.item[Y] < (-1.*s->width/2) && pt2.item[Y] < (-1.*s->width/2) ) ) {
		return 0;
	}
	
	/* See if the whole segement can fit into one point, if so, then just draw the point! */
	if ( (int)pt1.item[X] == (int)pt2.item[X] && (int)pt1.item[Y] == (int)pt2.item[Y] ) {
		return POINT_ONLY;
	}

	return 1;
}


void draw_circles(struct imageLayout *s) {
	int i,j; 
	spherePt spt1,spt2;
	vector pt1, pt2;	
	int inc=15; //draw line ever so many degrees

	spt1.item[RADIUS] = EARTH_RADIUS;
	spt2.item[RADIUS] = EARTH_RADIUS;

	for (i=0; i<360; i+=inc) {
		for (j=0; j<360; j+=5) {
			spt1.item[LONGITUDE] = i;
			spt1.item[LATITUDE]  = j;
			spt2.item[LONGITUDE] = i;
			spt2.item[LATITUDE]  = j+5;
			pt1 = multiply23(sphere2cart(spt1), s->rotation);
			pt2 = multiply23(sphere2cart(spt2), s->rotation);
			if (pt1.item[Z] > 0) 
				drawLine(s, pt1, pt2, s->colormap[CONF][C_LAT_GRID]);
		}
	}
	for (i=0; i<360; i+=inc) {
		for (j=0; j<360; j+=5) {
			spt1.item[LONGITUDE] = j;
			spt1.item[LATITUDE]  = i;
			spt2.item[LONGITUDE] = j+5;
			spt2.item[LATITUDE]  = i;
			pt1 = multiply23(sphere2cart(spt1), s->rotation);
			pt2 = multiply23(sphere2cart(spt2), s->rotation);
			if (pt1.item[Z] > 0) 
				drawLine(s, pt1, pt2, s->colormap[CONF][C_LONG_GRID]);
		}
	}
}
		
		
void addText (struct imageLayout *s, spherePt spt, char* text) {
	vector pt;
	int x,y;
	int text_x, text_y;
	int circle_size = CIRCLE_SIZE;
	
	//determine the point, post transformation
	pt = multiply23(sphere2cart(spt), s->rotation);

	if (pt.item[Z] < 0) 
		return;

	x = pt.item[X] + s->width/2;
	y = pt.item[Y] + s->height/2;

	//verify that the point is on the map, return if not
	if (x > s->width || x < 0 || y > s->height || y < 0)
		return;
	
	// draw a circle around the point in question
	gdImageArc(s->im, x,y, circle_size, circle_size, 0,360, s->colormap[CONF][C_TEXT_CIRCLE]);

	// determine the text location relative to the circle
	text_x = (x > s->width/2) ? (x - strlen(text)*gdFontSmall->w - circle_size/2) : 
				  (x + circle_size/2);
	text_y = (y > s->height/2) ? (y - circle_size/2) : 
				   (y + gdFontSmall->h + circle_size/2);
	text_y = y;
	// add text
	gdImageString(s->im, gdFontSmall, text_x, text_y, text, s->colormap[CONF][C_TEXT]);
}

void parse_cityfile (struct imageLayout *s) {
	FILE *f;	
	char string[256];
	int ret, line_number=0;
	int _long, _lat, i, j, pos;
	char _city_name[100];
	spherePt sp;

	if (s->o.cityfile == NULL) 
		return;

	sp.item[RADIUS] = EARTH_RADIUS;

	f = fopen(s->o.cityfile, "r");
	if (f == NULL) {
		fprintf(stderr,"Couldn't open %s (cityfile)\n", s->o.cityfile);
		exit(-2);
	}
	
	while ((ret = fread(string, 1, 255, f)) != 0) {
		line_number++;
		pos = (int) ftell(f) - ret;
		for (i=0; i<255; i++) {
			if (string[i] == '\n') {
				string[i] = '\0';
				fseek(f,pos+i+1, SEEK_SET);
				break;
			}
		}
		/* Remove any comments */
		for (j=0; j<i; j++) {
			if (string[j] == '#')  {
				string[j] = '\0';
				break;
			}
		} 
		/* remove any whitespaces before the comment */
		while (j>0) {
			j--;
			if (string[j] == ' ' || string [j] == '\t') {
				string[j] = '\0';
			} else {
				break;
			}
		}
		/* if the line is blank, skip it */
		if (strlen(string) == 0) 
			continue;

		ret = sscanf(string, "%d %d %s", &_long, &_lat, _city_name);
		if (ret != 3) {
			fprintf(stderr,"Unable to parse the string [%s] on line %d in %s\n", 
				string,line_number,s->o.cityfile);	
			continue;
		}
		if (_long > 648000 || _long < -648000 || _lat > 324000 || _lat < -324000) {
			fprintf(stderr,"Declaration value out of range (long=%d,lat=%d) on line %d in %s\n",
				_long, _lat, line_number, s->o.colorfile);
		} else {
			sp.item[LONGITUDE] = _long/3600.0;
			sp.item[LATITUDE] = _lat/3600.0;
			addText(s, sp, _city_name);
		}
	}
	fclose(f);
}
/* This function will generate the proper transformation 
 * matrix (rotation and scaling) and copy the values 
 * to the imageLayout struct
 */

void computeMatrix(struct imageLayout *s) {
	int i,j;
	matrix rot;
	// generate the proper rotation&scale matrix
	rot = multiply33 (
		multiply33( 
			multiply33( 
				rotateZ(180), 
				multiply33( 
					rotateY(s->o.yrot), 
					rotateX(s->o.xrot)
					)
				), 
			scale(s->o.zoom, s->o.zoom, s->o.zoom)
		),
		rotateZ(s->o.zrot)
		);

	// copy the matrix to the imageLayout struct
	for (i=0; i<3; i++)
		for (j=0; j<3; j++)
			s->rotation.item[i][j] = rot.item[i][j];
}
	
void generate(struct imageLayout *scene) {
	// misc variables
	spherePt s1,s2;
	vector pt1, pt2;
	int i,j,ret;
	//filehandle
	int fh; 
	//filenames
	struct header header_data;
	struct segment_index segment_index_data;
	struct segment_header segment_header_data;
	struct stroke stroke_data;
	//simple mapping array
	int bit_number[9] = {0,1,2,0,3,0,0,0,4};

	// parse the colors from a file if provided
	// otherwise, assume it's been already setup before this subroutine
	// has been called
	if (scene->o.colorfile) 
		parse_colorfile(scene);

	if (scene->o.cityfile) 
		parse_cityfile(scene);

	if (scene->o.gridlines) 
		draw_circles(scene);
	
	// iterate through data file and add the cities if appropriate
	if (scene->o.cities) {
		i = 0;
		s1.item[RADIUS]=EARTH_RADIUS;
		while (cities[i].name != NULL) {
			s1.item[LONGITUDE] = cities[i].longitude;
			s1.item[LATITUDE] = cities[i].latitude;
			addText(scene, s1, cities[i].name);
			i++;
		}
	}

#ifdef DEBUG
	printf("-- GENERATE --\n");
	printf("o Options\n");
	printf("  - cities    [%d]\n", scene->o.cities);
	printf("  - gridlines [%d]\n", scene->o.gridlines);
	printf("  - zoom      [%g]\n", scene->o.zoom);
	printf("  - xrot      [%g]\n", scene->o.xrot);
	printf("  - yrot      [%g]\n", scene->o.yrot);
	printf("  - zrot      [%g]\n", scene->o.zrot);
	printf("  - zrot      [%g]\n", scene->o.zrot);
	printf("  - mapfile   [%s]\n", scene->o.mapfile);
	printf("  - outfile   [%s]\n", scene->o.outfile);
	printf("  - colorfile [%s]\n", scene->o.colorfile);
	printf("  - des_cat   [%X]\n", scene->o.desired_categories);
	printf("  - des_cont  [%X]\n", scene->o.desired_continents);
	printf("o imageLayout (Scene)\n");
	printf("  - height    [%d]\n", scene->height);
	printf("  - width     [%d]\n", scene->width);
	printf("  - transp.   [%d]\n", scene->transparent);
	printf("  - gdImgPtr  [%X]\n", (unsigned int) scene->im);
	printf("  o Colormap\n");
	for (i=0; i<5; i++) 
		for (j=0; j<16; j++) 
			if (scene->colormap[i][j] != NOPRINT)
				printf("    - (%2d,%2d) [%d]\n",i ,j ,scene->colormap[i][j]);
	printf("  o Rotation Matrix\n");
	for (i=0; i<3; i++) 
		printf("     [%e][%e][%e]\n", scene->rotation.item[i][0],  
			scene->rotation.item[i][1], scene->rotation.item[i][2]);
	printf("-- -------- --\n");
#endif DEBUG

	//Open the datafile and find graph the appropriate sectors
	fh = open(scene->o.mapfile, O_RDONLY, 0);
	if (fh == -1) {
		fprintf(stderr,"Couldn't open input vector file %s.\n", scene->o.mapfile);
		exit (-1);
	}
	// read the header
	read(fh, &header_data, sizeof(header_data));
#ifdef WORDS_BIGENDIAN
	bitflip( &(header_data.magic) );	
	bitflip( &(header_data.segment_index_address) );	
	bitflip( &(header_data.segment_index_count) );	
#endif /* WORDS_BIGENDIAN */
	if (header_data.magic != HEADER_MAGIC) {
		fprintf(stderr, "The magic was not valid for the input file.\n");
		fprintf(stderr, "This could be an endian problem, or you\n"); 
		fprintf(stderr, "could be pointing to the wrong file.\n");
		fprintf(stderr, "(%s's magic = %X and it's supposed to be %X)\n", 
			scene->o.mapfile, header_data.magic, HEADER_MAGIC);
		exit(-2);
	}
	// itterate through all the segment_indexes to see which ones are interesting
	for (i=0; i<header_data.segment_index_count; i++) {
		lseek(fh, sizeof(header_data) + sizeof(segment_index_data)*i, SEEK_SET);
		read(fh, &segment_index_data, sizeof(segment_index_data));
#ifdef WORDS_BIGENDIAN
		bitflip( &(segment_index_data.maxlat) );
		bitflip( &(segment_index_data.minlat) );
		bitflip( &(segment_index_data.maxlong) );
		bitflip( &(segment_index_data.minlong) );
		bitflip( &(segment_index_data.segment_address) );
		bitflip( &(segment_index_data.continent) );
		bitflip( &(segment_index_data.category) );
		bitflip( &(segment_index_data.type) );
#endif /* WORDS_BIGENDIAN */

		if ( (segment_index_data.continent & scene->o.desired_continents)
		 && (segment_index_data.category & scene->o.desired_categories) ) { 
			//it has interesting information, let's look to see if the
			//data is in our viewport
			ret = segment_in_scene(scene, &segment_index_data);
			if (ret == YES) {
				//iterate through the data and pass it to draw
				lseek(fh, segment_index_data.segment_address, SEEK_SET);
				read(fh, &segment_header_data, sizeof(segment_header_data));
#ifdef WORDS_BIGENDIAN
				bitflip( &(segment_header_data.orgx) );
				bitflip( &(segment_header_data.orgy) );
				bitflip( &(segment_header_data.nstrokes) );
#endif /* WORDS_BIGENDIAN */
				s1.item[LONGITUDE] = (segment_header_data.orgx/3600.0);
				s1.item[LATITUDE]  = (segment_header_data.orgy/3600.0);
				s1.item[RADIUS]=EARTH_RADIUS;
				s2.item[RADIUS]=EARTH_RADIUS;
				for (j=0; j<segment_header_data.nstrokes; j++) {
					// itterate and draw all the appropriate points
					read(fh, &stroke_data, sizeof(stroke_data));
#ifdef WORDS_BIGENDIAN
					bitflip( &(stroke_data.dx) );
					bitflip( &(stroke_data.dy) );
#endif /* WORDS_BIGENDIAN */
					s2.item[LONGITUDE] = s1.item[LONGITUDE]+(stroke_data.dx/3600.0);
					s2.item[LATITUDE]  = s1.item[LATITUDE] +(stroke_data.dy/3600.0);

					pt1 = multiply23(sphere2cart(s1), scene->rotation);
					pt2 = multiply23(sphere2cart(s2), scene->rotation);
	
					drawLine(scene, pt1, pt2, 
						scene->colormap[bit_number[segment_index_data.category]][segment_index_data.type]);

					s1.item[LONGITUDE] = s2.item[LONGITUDE];
					s1.item[LATITUDE]  = s2.item[LATITUDE];
				}
			} else if (ret == POINT_ONLY) {
				// the whole segment amounts to a point based on the
				// perspective of the scene, so just draw a point	
				s1.item[LONGITUDE] = segment_index_data.maxlong/3600.;
				s1.item[LATITUDE] = segment_index_data.maxlat/3600.;
				s1.item[RADIUS] = EARTH_RADIUS;
				pt1 = multiply23(sphere2cart(s1), scene->rotation);
				drawPoint(scene, pt1, 
					scene->colormap[bit_number[segment_index_data.category]][segment_index_data.type]);
			}
		}
	}

	makePicture(scene);
}

